package dk.camelot64.kickc.icl;

import java.util.ArrayList;
import java.util.List;

/**
 * A live range equivalence class contains a set of variables with non-overlapping live ranges.
 **/
public class LiveRangeEquivalenceClass {

   /** The containing program. */
   private Program program;

   /** The variables of the equivalence class. */
   private List<VariableRef> variables;

   /** The combined live range of the variables. */
   private LiveRange classLiveRange;

   public LiveRangeEquivalenceClass(Program program) {
      this.program = program;
      this.variables = new ArrayList<>();
      this.classLiveRange = new LiveRange();
   }

   /**
    * Add a variable to the equivalence class
    *
    * @param variable The variable to add
    */
   public void addVariable(VariableRef variable) {
      if(variables.contains(variable)) {
         return;
      }
      LiveRangeVariables liveRanges = program.getScope().getLiveRanges();
      LiveRange varLiveRange = liveRanges.getLiveRange(variable);
      if (classLiveRange.overlaps(varLiveRange)) {
         throw new RuntimeException("Compilation error! Variable live range overlaps live range equivalence class live range. " + variable);
      }
      classLiveRange.add(varLiveRange);
      variables.add(variable);
   }

   /**
    * Determines if the phi equivalence class contains a variable
    *
    * @param variable The variable to look for
    * @return true if the equivalence class contains the variable
    */
   public boolean contains(VariableRef variable) {
      return variables.contains(variable);
   }

   public LiveRange getLiveRange() {
      return classLiveRange;
   }

   public void addAll(LiveRangeEquivalenceClass other) {
      variables.addAll(other.variables);
      classLiveRange.add(other.classLiveRange);
   }

   @Override
   public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      LiveRangeEquivalenceClass that = (LiveRangeEquivalenceClass) o;

      return variables.equals(that.variables);
   }

   @Override
   public int hashCode() {
      return variables.hashCode();
   }

   @Override
   public String toString() {
      StringBuilder s = new StringBuilder();
      s.append("[ ");
      for (VariableRef variable : variables) {
         s.append(variable.toString());
         s.append(" ");
      }
      s.append("]");

      return s.toString();
   }

   public List<VariableRef> getVariables() {
      return variables;
   }
}
